@@ -24,6 +24,7 @@ raw data specified in version discovery responses.
 import copy
 import re
 
+import os_service_types
 import six
 from six.moves import urllib
 
@@ -34,6 +35,7 @@ from keystoneauth1 import exceptions
 _LOGGER = utils.get_logger(__name__)
 
 LATEST = float('inf')
+_SERVICE_TYPES = os_service_types.ServiceTypes()
 
 
 def _str_or_latest(val):
@@ -218,7 +220,28 @@ def normalize_version_number(version):
     raise TypeError('Invalid version specified: %s' % version)
 
 
-def _normalize_version_args(version, min_version, max_version):
+def _normalize_version_args(
+        version, min_version, max_version, service_type=None):
+    # The sins of our fathers become the blood on our hands.
+    # If a user requests an old-style service type such as volumev2, then they
+    # are inherently requesting the major API version 2. It's not a good
+    # interface, but it's the one that was imposed on the world years ago
+    # because the client libraries all hid the version discovery document.
+    # In order to be able to ensure that a user who requests volumev2 does not
+    # get a block-storage endpoint that only provides v3 of the block-storage
+    # service, we need to pull the version out of the service_type. The
+    # service-types-authority will prevent the growth of new monstrosities such
+    # as this, but in order to move forward without breaking people, we have
+    # to just cry in the corner while striking ourselves with thorned branches.
+    # That said, for sure only do this hack for officially known service_types.
+    if (service_type and
+            _SERVICE_TYPES.is_known(service_type) and
+            service_type[-1].isdigit() and
+            service_type[-2] == 'v'):
+        implied_version = normalize_version_number(service_type[-1])
+    else:
+        implied_version = None
+
     if version and (min_version or max_version):
         raise ValueError(
             "version is mutually exclusive with min_version and max_version")
@@ -227,6 +250,12 @@ def _normalize_version_args(version, min_version, max_version):
         # Explode this into min_version and max_version
         min_version = normalize_version_number(version)
         max_version = (min_version[0], LATEST)
+        if implied_version:
+            if min_version[0] != implied_version[0]:
+                raise exceptions.ImpliedVersionMismatch(
+                    service_type=service_type,
+                    implied=implied_version,
+                    given=version_to_string(version))
         return min_version, max_version
 
     if min_version == 'latest':
@@ -255,6 +284,27 @@ def _normalize_version_args(version, min_version, max_version):
     if None not in (min_version, max_version) and max_version < min_version:
         raise ValueError("min_version cannot be greater than max_version")
 
+    if implied_version:
+        if min_version:
+            if min_version[0] != implied_version[0]:
+                raise exceptions.ImpliedMinVersionMismatch(
+                    service_type=service_type,
+                    implied=implied_version,
+                    given=version_to_string(min_version))
+        else:
+            min_version = implied_version
+
+        # If 'latest' is provided with a versioned service-type like
+        # volumev2 - the user wants the latest of volumev2, not the latest
+        # of block-storage.
+        if max_version and max_version[0] != LATEST:
+            if max_version[0] != implied_version[0]:
+                raise exceptions.ImpliedMaxVersionMismatch(
+                    service_type=service_type,
+                    implied=implied_version,
+                    given=version_to_string(max_version))
+        else:
+            max_version = (implied_version[0], LATEST)
     return min_version, max_version
 
 
@@ -389,6 +439,35 @@ def _version_from_url(url):
     return None
 
 
+class Status(object):
+    CURRENT = 'CURRENT'
+    SUPPORTED = 'SUPPORTED'
+    DEPRECATED = 'DEPRECATED'
+    EXPERIMENTAL = 'EXPERIMENTAL'
+    UNKNOWN = 'UNKNOWN'
+    KNOWN = (CURRENT, SUPPORTED, DEPRECATED, EXPERIMENTAL)
+
+    @classmethod
+    def normalize(cls, raw_status):
+        """Turn a status into a canonical status value.
+
+        If the status from the version discovery document does not match one
+        of the known values, it will be set to 'UNKNOWN'.
+
+        :param str raw_status: Status value from a discovery document.
+
+        :returns: A canonicalized version of the status. Valid values
+                  are CURRENT, SUPPORTED, DEPRECATED, EXPERIMENTAL and UNKNOWN
+        :rtype: str
+        """
+        status = raw_status.upper()
+        if status == 'STABLE':
+            status = 'CURRENT'
+        if status not in cls.KNOWN:
+            status = cls.UNKNOWN
+        return status
+
+
 class Discover(object):
 
     CURRENT_STATUSES = ('stable', 'current', 'supported')
@@ -448,19 +527,8 @@ class Discover(object):
         :param bool reverse: Reverse the list. reverse=true will mean the
                              returned list is sorted from newest to oldest
                              version.
-        :returns: A list of version data dictionaries sorted by version number.
-                  Each data element in the returned list is a dictionary
-                  consisting of:
-
-          :version tuple: The normalized version of the endpoint.
-          :url str: The url for the endpoint.
-          :collection: The URL for the discovery document.  May be None.
-          :min_microversion: The minimum microversion supported by the
-                             endpoint.  May be None.
-          :max_microversion: The maximum microversion supported by the
-                             endpoint.  May be None.
-          :raw_status str: The status as provided by the server
-        :rtype: list(dict)
+        :returns: A list of :class:`VersionData` sorted by version number.
+        :rtype: list(VersionData)
         """
         data = self.raw_version_data(**kwargs)
         versions = []
@@ -518,18 +586,38 @@ class Discover(object):
                              'Missing link to endpoint.')
                 continue
 
-            versions.append({'version': version_number,
-                             'url': self_url,
-                             'collection': collection_url,
-                             'min_microversion': min_microversion,
-                             'max_microversion': max_microversion,
-                             'next_min_version': next_min_version,
-                             'not_before': not_before,
-                             'raw_status': v['status']})
+            versions.append(
+                VersionData(version=version_number,
+                            url=self_url,
+                            collection=collection_url,
+                            min_microversion=min_microversion,
+                            max_microversion=max_microversion,
+                            next_min_version=next_min_version,
+                            not_before=not_before,
+                            status=Status.normalize(v['status']),
+                            raw_status=v['status']))
 
         versions.sort(key=lambda v: v['version'], reverse=reverse)
         return versions
 
+    def version_string_data(self, reverse=False, **kwargs):
+        """Get normalized version data with versions as strings.
+
+        Return version data in a structured way.
+
+        :param bool reverse: Reverse the list. reverse=true will mean the
+                             returned list is sorted from newest to oldest
+                             version.
+        :returns: A list of :class:`VersionData` sorted by version number.
+        :rtype: list(VersionData)
+        """
+        version_data = self.version_data(reverse=reverse, **kwargs)
+        for version in version_data:
+            for key in ('version', 'min_microversion', 'max_microversion'):
+                if version[key]:
+                    version[key] = version_to_string(version[key])
+        return version_data
+
     def data_for(self, version, **kwargs):
         """Return endpoint data for a version.
 
@@ -662,6 +750,79 @@ class Discover(object):
         return data['url'] if data else None
 
 
+class VersionData(dict):
+    """Normalized Version Data about an endpoint."""
+
+    def __init__(
+            self,
+            version,
+            url,
+            collection=None,
+            max_microversion=None,
+            min_microversion=None,
+            next_min_version=None,
+            not_before=None,
+            status='CURRENT',
+            raw_status=None):
+        super(VersionData, self).__init__()
+        self['version'] = version
+        self['url'] = url
+        self['collection'] = collection
+        self['max_microversion'] = max_microversion
+        self['min_microversion'] = min_microversion
+        self['next_min_version'] = next_min_version
+        self['not_before'] = not_before
+        self['status'] = status
+        self['raw_status'] = raw_status
+
+    @property
+    def version(self):
+        """The normalized version of the endpoint."""
+        return self.get('version')
+
+    @property
+    def url(self):
+        """The url for the endpoint."""
+        return self.get('url')
+
+    @property
+    def collection(self):
+        """The URL for the discovery document.
+
+        May be None.
+        """
+        return self.get('collection')
+
+    @property
+    def min_microversion(self):
+        """The minimum microversion supported by the endpoint.
+
+        May be None.
+        """
+        return self.get('min_microversion')
+
+    @property
+    def max_microversion(self):
+        """The maximum microversion supported by the endpoint.
+
+        May be None.
+        """
+        return self.get('max_microversion')
+
+    @property
+    def status(self):
+        """A canonicalized version of the status.
+
+        Valid values are CURRENT, SUPPORTED, DEPRECATED and EXPERIMENTAL.
+        """
+        return self.get('status')
+
+    @property
+    def raw_status(self):
+        """The status as provided by the server."""
+        return self.get('raw_status')
+
+
 class EndpointData(object):
     """Normalized information about a discovered endpoint.
 
@@ -688,7 +849,8 @@ class EndpointData(object):
                  min_microversion=None,
                  max_microversion=None,
                  next_min_version=None,
-                 not_before=None):
+                 not_before=None,
+                 status=None):
         self.catalog_url = catalog_url
         self.service_url = service_url
         self.service_type = service_type
@@ -703,6 +865,7 @@ class EndpointData(object):
         self.max_microversion = max_microversion
         self.next_min_version = next_min_version
         self.not_before = not_before
+        self.status = status
         self._saved_project_id = None
         self._catalog_matches_version = False
         self._catalog_matches_exactly = False
@@ -727,7 +890,9 @@ class EndpointData(object):
             min_microversion=self.min_microversion,
             max_microversion=self.max_microversion,
             next_min_version=self.next_min_version,
-            not_before=self.not_before)
+            not_before=self.not_before,
+            status=self.status,
+        )
         # Save cached discovery object - but we don't want to
         # actually provide a constructor argument
         new_data._disc = self._disc
@@ -840,6 +1005,50 @@ class EndpointData(object):
             max_version=max_version)
         return new_data
 
+    def get_all_version_string_data(self, session, project_id=None):
+        """Return version data for all versions discovery can find.
+
+        :param string project_id: ID of the currently scoped project. Used for
+                                  removing project_id components of URLs from
+                                  the catalog. (optional)
+        :returns: A list of :class:`VersionData` sorted by version number.
+        :rtype: list(VersionData)
+        """
+        versions = []
+        for vers_url in self._get_discovery_url_choices(project_id=project_id):
+            try:
+                d = get_discovery(session, vers_url)
+            except Exception as e:
+                # Ignore errors here - we're just searching for one of the
+                # URLs that will give us data.
+                _LOGGER.debug(
+                    "Failed attempt at discovery on %s: %s", vers_url, str(e))
+                continue
+            for version in d.version_string_data():
+                versions.append(version)
+            break
+        return versions or self._infer_version_data(project_id)
+
+    def _infer_version_data(self, project_id=None):
+        """Return version data dict for when discovery fails.
+
+        :param string project_id: ID of the currently scoped project. Used for
+                                  removing project_id components of URLs from
+                                  the catalog. (optional)
+        :returns: A list of :class:`VersionData` sorted by version number.
+        :rtype: list(VersionData)
+        """
+        version = self.api_version
+        if version:
+            version = version_to_string(self.api_version)
+
+        url = self.url.rstrip("/")
+        if project_id and url.endswith(project_id):
+            url, _ = self.url.rsplit('/', 1)
+        url += "/"
+
+        return [VersionData(url=url, version=version)]
+
     def _set_version_info(self, session, allow=None, cache=None,
                           allow_version_hack=True, project_id=None,
                           discover_versions=False,
@@ -900,6 +1109,7 @@ class EndpointData(object):
         self.next_min_version = discovered_data['next_min_version']
         self.not_before = discovered_data['not_before']
         self.api_version = discovered_data['version']
+        self.status = discovered_data['status']
 
         # TODO(mordred): these next two things should be done by Discover
         # in versioned_data_for.
